import json

import httpx
import pytest
import respx

from unittest.mock import MagicMock, patch

from mrp.agent import Agent
from mrp.auth import Keypair, _b64url_decode
from mrp.crypto import encrypt_message, decrypt_message
from mrp.errors import RequestError
from mrp.models import Message
from mrp.schemas import REQUEST_CONTENT_TYPE, RESPONSE_CONTENT_TYPE, create_request


@pytest.fixture
def keypair():
    return Keypair.generate()


def _mock_register(keypair):
    """Set up mock for agent registration (PATCH)."""
    pk = keypair.public_key_b64
    respx.patch(f"http://localhost:8080/v1/agents/{pk}").mock(
        return_value=httpx.Response(200, json={
            "public_key": pk, "display_name": "Test", "status": "active",
            "metadata": {}, "capabilities": [],
            "created_at": "2026-03-05T12:00:00Z",
            "last_active_at": "2026-03-05T12:00:00Z",
        })
    )


def _make_request_message(keypair, action="echo", params=None):
    """Create a Message with REQUEST_CONTENT_TYPE for dispatch testing."""
    if params is None:
        params = {"text": "hello"}
    req = create_request(action, params)
    return Message(
        message_id="msg_1",
        sender_key="sender_key_abc",
        recipient_key=keypair.public_key_b64,
        content_type=REQUEST_CONTENT_TYPE,
        body=req.to_dict(),
        attachments=[],
        status="delivered",
        thread_id="t1",
        in_reply_to=None,
        metadata={},
        created_at="2026-03-05T12:00:00Z",
        expires_at="2026-03-12T12:00:00Z",
    )


def _make_plain_message(keypair):
    """Create a plain application/json Message for dispatch testing."""
    return Message(
        message_id="msg_2",
        sender_key="sender_key_abc",
        recipient_key=keypair.public_key_b64,
        content_type="application/json",
        body={"text": "hello"},
        attachments=[],
        status="delivered",
        thread_id=None,
        in_reply_to=None,
        metadata={},
        created_at="2026-03-05T12:00:00Z",
        expires_at="2026-03-12T12:00:00Z",
    )


class TestAgentDispatch:
    @respx.mock
    def test_action_routing(self, keypair):
        """Register handler with on_request, dispatch request message, verify handler called."""
        agent = Agent("http://localhost:8080", keypair=keypair)
        _mock_register(keypair)

        # Mock the respond call (it sends an HTTP message)
        respx.post("http://localhost:8080/v1/messages").mock(
            return_value=httpx.Response(202, json={
                "message_id": "msg_reply", "status": "sent",
                "created_at": "2026-03-05T12:00:00Z",
                "expires_at": "2026-03-12T12:00:00Z",
            })
        )

        handler_called_with = []

        def echo_handler(req, msg):
            handler_called_with.append((req, msg))
            return {"echo": req.params["text"]}

        agent.on_request("echo", echo_handler)

        msg = _make_request_message(keypair, action="echo", params={"text": "hello"})
        result = agent._dispatch_message(msg)

        # Action handler returns None (auto-responds via respond())
        assert result is None
        assert len(handler_called_with) == 1
        assert handler_called_with[0][0].action == "echo"
        assert handler_called_with[0][0].params == {"text": "hello"}
        assert handler_called_with[0][1].message_id == "msg_1"

    @respx.mock
    def test_action_error_response(self, keypair):
        """Handler raises RequestError -- verify respond_error is called."""
        agent = Agent("http://localhost:8080", keypair=keypair)
        _mock_register(keypair)

        # Capture the sent response
        route = respx.post("http://localhost:8080/v1/messages").mock(
            return_value=httpx.Response(202, json={
                "message_id": "msg_err", "status": "sent",
                "created_at": "2026-03-05T12:00:00Z",
                "expires_at": "2026-03-12T12:00:00Z",
            })
        )

        def failing_handler(req, msg):
            raise RequestError("NOT_FOUND", "resource not found")

        agent.on_request("lookup", failing_handler)

        msg = _make_request_message(keypair, action="lookup", params={"id": "123"})
        result = agent._dispatch_message(msg)

        assert result is None
        # Verify respond_error sent a message
        assert route.called
        sent_body = json.loads(route.calls.last.request.content)
        assert sent_body["content_type"] == RESPONSE_CONTENT_TYPE
        body = sent_body["body"]
        assert body["status"] == "error"
        assert body["error"]["code"] == "NOT_FOUND"
        assert body["error"]["message"] == "resource not found"

    def test_fallback_on_message(self, keypair):
        """Set on_message, dispatch non-request message, verify called."""
        agent = Agent("http://localhost:8080", keypair=keypair)

        handler_called_with = []

        def generic_handler(msg):
            handler_called_with.append(msg)
            return {"reply": "ok"}

        agent.on_message(generic_handler)

        msg = _make_plain_message(keypair)
        result = agent._dispatch_message(msg)

        assert len(handler_called_with) == 1
        assert handler_called_with[0].message_id == "msg_2"
        assert result == {"reply": "ok"}


class TestAgentEncryption:
    @respx.mock
    def test_send_encrypted(self, keypair):
        """Send with encrypt=True, verify content_type changed to encrypted."""
        agent = Agent("http://localhost:8080", keypair=keypair)
        _mock_register(keypair)

        # Create a recipient keypair
        recipient = Keypair.generate()

        route = respx.post("http://localhost:8080/v1/messages").mock(
            return_value=httpx.Response(202, json={
                "message_id": "msg_enc", "status": "sent",
                "created_at": "2026-03-05T12:00:00Z",
                "expires_at": "2026-03-12T12:00:00Z",
            })
        )

        result = agent.send(
            to=recipient.public_key_b64,
            body={"secret": "data"},
            encrypt=True,
        )

        assert result.message_id == "msg_enc"
        assert route.called

        sent_body = json.loads(route.calls.last.request.content)
        assert sent_body["content_type"] == "application/x-m2m-encrypted"
        # The body should have encrypted fields
        encrypted_body = sent_body["body"]
        assert encrypted_body["v"] == 2
        assert "enc" in encrypted_body
        assert "ct" in encrypted_body
        assert encrypted_body["ct_content_type"] == "application/json"

    def test_decrypt_success(self, keypair):
        """Encrypt then decrypt roundtrip via Agent.decrypt()."""
        sender = Keypair.generate()
        recipient = Keypair.generate()

        # Encrypt a message body
        plaintext = json.dumps({"secret": "hello"}).encode()
        encrypted_body = encrypt_message(
            sender.seed,
            recipient.public_key,
            plaintext,
            "application/json",
        )

        # Create a Message with encrypted content
        msg = Message(
            message_id="msg_enc",
            sender_key=sender.public_key_b64,
            recipient_key=recipient.public_key_b64,
            content_type="application/x-m2m-encrypted",
            body=encrypted_body,
            attachments=[],
            status="delivered",
            thread_id=None,
            in_reply_to=None,
            metadata={},
            created_at="2026-03-05T12:00:00Z",
            expires_at="2026-03-12T12:00:00Z",
        )

        # Decrypt via Agent
        agent = Agent("http://localhost:8080", keypair=recipient)
        decrypted_bytes, content_type = agent.decrypt(msg)

        assert json.loads(decrypted_bytes) == {"secret": "hello"}
        assert content_type == "application/json"

    def test_decrypt_not_encrypted(self, keypair):
        """ValueError for non-encrypted content type."""
        agent = Agent("http://localhost:8080", keypair=keypair)

        msg = Message(
            message_id="msg_plain",
            sender_key="sender_key",
            recipient_key=keypair.public_key_b64,
            content_type="application/json",
            body={"text": "hello"},
            attachments=[],
            status="delivered",
            thread_id=None,
            in_reply_to=None,
            metadata={},
            created_at="2026-03-05T12:00:00Z",
            expires_at="2026-03-12T12:00:00Z",
        )

        with pytest.raises(ValueError, match="Message is not encrypted"):
            agent.decrypt(msg)



"""Local contact store for mapping friendly names to public keys.

Contacts are stored as JSON at ~/.mrp/contacts.json and shared across
the CLI, MCP server, and OpenClaw plugin.
"""

from __future__ import annotations

import json
import os
from dataclasses import dataclass
from pathlib import Path


@dataclass
class Contact:
    """A saved contact mapping a friendly name to a public key."""

    name: str
    public_key: str
    note: str | None = None


class ContactStore:
    """Manages a local contact list persisted as JSON.

    The contact file is shared with the MRP CLI at ``~/.mrp/contacts.json``.
    """

    def __init__(self, path: str, contacts: list[Contact] | None = None) -> None:
        self._path = path
        self._contacts: list[Contact] = contacts or []

    @staticmethod
    def default_path() -> str:
        """Return the default contacts file path (~/.mrp/contacts.json)."""
        return os.path.join(Path.home(), ".mrp", "contacts.json")

    @classmethod
    def load(cls, path: str | None = None) -> ContactStore:
        """Load contacts from *path* (or the default).

        Returns an empty store if the file does not exist.
        """
        p = path or cls.default_path()
        try:
            data = json.loads(Path(p).read_text(encoding="utf-8"))
        except FileNotFoundError:
            return cls(p)
        except (json.JSONDecodeError, OSError) as exc:
            raise ValueError(f"parsing contacts: {exc}") from exc

        contacts: list[Contact] = []
        for entry in data.get("contacts", []):
            contacts.append(
                Contact(
                    name=entry["name"],
                    public_key=entry["public_key"],
                    note=entry.get("note"),
                )
            )
        return cls(p, contacts)

    def save(self) -> None:
        """Write the contact store to disk."""
        path = Path(self._path)
        path.parent.mkdir(parents=True, mode=0o700, exist_ok=True)
        entries = []
        for c in self._contacts:
            entry: dict[str, str] = {"name": c.name, "public_key": c.public_key}
            if c.note:
                entry["note"] = c.note
            entries.append(entry)
        data = json.dumps({"contacts": entries}, indent=2) + "\n"
        path.write_text(data, encoding="utf-8")

    def add(self, name: str, public_key: str, note: str | None = None) -> None:
        """Add a contact. Raises ``ValueError`` on duplicates or invalid names."""
        _validate_contact_name(name)
        lower = name.lower()
        for c in self._contacts:
            if c.name.lower() == lower:
                raise ValueError(f'contact "{c.name}" already exists')
            if c.public_key == public_key:
                raise ValueError(f'public key already saved as "{c.name}"')
        self._contacts.append(Contact(name=name, public_key=public_key, note=note))

    def remove(self, name: str) -> None:
        """Remove a contact by name (case-insensitive)."""
        lower = name.lower()
        for i, c in enumerate(self._contacts):
            if c.name.lower() == lower:
                del self._contacts[i]
                return
        raise ValueError(f'contact "{name}" not found')

    def resolve(self, name_or_key: str) -> str:
        """Map a name or key to a public key.

        If *name_or_key* contains ``:``, it is treated as a raw key and
        returned as-is.  Otherwise a case-insensitive name lookup is
        performed.  If no match is found the original input is returned
        unchanged.
        """
        if ":" in name_or_key:
            return name_or_key
        lower = name_or_key.lower()
        for c in self._contacts:
            if c.name.lower() == lower:
                return c.public_key
        return name_or_key

    def list(self) -> list[Contact]:
        """Return all contacts."""
        return list(self._contacts)

    def get(self, name: str) -> Contact | None:
        """Get a contact by name (case-insensitive), or ``None``."""
        lower = name.lower()
        for c in self._contacts:
            if c.name.lower() == lower:
                return c
        return None


def _validate_contact_name(name: str) -> None:
    if not name:
        raise ValueError("contact name must not be empty")
    if len(name) > 64:
        raise ValueError("contact name must be at most 64 characters")
    if ":" in name:
        raise ValueError("contact name must not contain ':'")



from __future__ import annotations

from mrp.auth import _b64url_decode, _b64url_encode

# These imports are optional - E2E encryption requires PyNaCl
try:
    import nacl.bindings
    _HAS_NACL = True
except ImportError:
    _HAS_NACL = False

# HPKE support requires pyhpke
try:
    from pyhpke import CipherSuite as HPKESuite, KEMId, KDFId, AEADId
    _HAS_HPKE = True
except ImportError:
    _HAS_HPKE = False

from cryptography.hazmat.primitives.asymmetric.x25519 import X25519PrivateKey, X25519PublicKey
from cryptography.hazmat.primitives import serialization


_HPKE_INFO_V2 = b"mrp-v2-msg"


def _ensure_nacl() -> None:
    if not _HAS_NACL:
        raise ImportError(
            "PyNaCl is required for E2E encryption. Install with: pip install mrp-sdk[e2e]"
        )


def _ensure_hpke() -> None:
    if not _HAS_HPKE:
        raise ImportError(
            "pyhpke is required for HPKE encryption. Install with: pip install mrp-sdk[e2e]"
        )


def _get_hpke_suite():
    _ensure_hpke()
    return HPKESuite.new(
        KEMId.DHKEM_X25519_HKDF_SHA256,
        KDFId.HKDF_SHA256,
        AEADId.CHACHA20_POLY1305,
    )


def ed25519_to_x25519_private(ed_seed: bytes) -> X25519PrivateKey:
    """Convert Ed25519 seed (32 bytes) to X25519 private key using nacl."""
    _ensure_nacl()
    # Use libsodium's conversion
    x_priv_bytes = nacl.bindings.crypto_sign_ed25519_sk_to_curve25519(ed_seed + b"\x00" * 32)
    return X25519PrivateKey.from_private_bytes(x_priv_bytes[:32])


def ed25519_to_x25519_public(ed_pub: bytes) -> X25519PublicKey:
    """Convert Ed25519 public key (32 bytes) to X25519 public key using nacl."""
    _ensure_nacl()
    x_pub_bytes = nacl.bindings.crypto_sign_ed25519_pk_to_curve25519(ed_pub)
    return X25519PublicKey.from_public_bytes(x_pub_bytes)


def encrypt_message(
    sender_seed: bytes,
    recipient_public_key: bytes,
    plaintext: bytes,
    content_type: str = "application/json",
) -> dict:
    """Encrypt a message for a recipient using HPKE Auth mode (v2).

    Args:
        sender_seed: 32-byte Ed25519 seed of the sender
        recipient_public_key: 32-byte Ed25519 public key of the recipient
        plaintext: message bytes to encrypt
        content_type: original content type

    Returns:
        dict with fields: v, enc, ct, ct_content_type
    """
    _ensure_nacl()
    _ensure_hpke()

    # Convert Ed25519 keys to X25519 raw bytes
    recipient_x25519 = ed25519_to_x25519_public(recipient_public_key)
    recipient_x25519_bytes = recipient_x25519.public_bytes(
        serialization.Encoding.Raw, serialization.PublicFormat.Raw
    )
    sender_x25519 = ed25519_to_x25519_private(sender_seed)
    sender_x25519_bytes = sender_x25519.private_bytes(
        serialization.Encoding.Raw, serialization.PrivateFormat.Raw, serialization.NoEncryption()
    )

    suite = _get_hpke_suite()
    pkr = suite.kem.deserialize_public_key(recipient_x25519_bytes)
    sks = suite.kem.deserialize_private_key(sender_x25519_bytes)

    enc, ctx = suite.create_sender_context(pkr=pkr, info=_HPKE_INFO_V2, sks=sks)
    ct = ctx.seal(plaintext)

    return {
        "v": 2,
        "enc": _b64url_encode(enc),
        "ct": _b64url_encode(ct),
        "ct_content_type": content_type,
    }


def decrypt_message(
    recipient_seed: bytes,
    sender_public_key: bytes,
    encrypted_body: dict,
) -> tuple[bytes, str]:
    """Decrypt an encrypted message.

    Args:
        recipient_seed: 32-byte Ed25519 seed of the recipient
        sender_public_key: 32-byte Ed25519 public key of the sender
        encrypted_body: dict with encryption fields

    Returns:
        (plaintext, original_content_type)
    """
    v = encrypted_body.get("v")
    if v != 2:
        raise ValueError(f"Unsupported encryption version: {v}")
    return _decrypt_v2(recipient_seed, sender_public_key, encrypted_body)


def _decrypt_v2(
    recipient_seed: bytes,
    sender_public_key: bytes,
    encrypted_body: dict,
) -> tuple[bytes, str]:
    """Decrypt a v2 encrypted message (HPKE Auth mode)."""
    _ensure_nacl()
    _ensure_hpke()

    enc = _b64url_decode(encrypted_body["enc"])
    ct = _b64url_decode(encrypted_body["ct"])
    content_type = encrypted_body["ct_content_type"]

    recipient_x25519 = ed25519_to_x25519_private(recipient_seed)
    recipient_x25519_bytes = recipient_x25519.private_bytes(
        serialization.Encoding.Raw, serialization.PrivateFormat.Raw, serialization.NoEncryption()
    )
    sender_x25519 = ed25519_to_x25519_public(sender_public_key)
    sender_x25519_bytes = sender_x25519.public_bytes(
        serialization.Encoding.Raw, serialization.PublicFormat.Raw
    )

    suite = _get_hpke_suite()
    skr = suite.kem.deserialize_private_key(recipient_x25519_bytes)
    pks = suite.kem.deserialize_public_key(sender_x25519_bytes)

    ctx = suite.create_recipient_context(enc=enc, skr=skr, info=_HPKE_INFO_V2, pks=pks)
    plaintext = ctx.open(ct)

    return plaintext, content_type



from __future__ import annotations

import asyncio
import inspect
import json
import logging
import time
from collections.abc import Iterator
from io import IOBase
from typing import Any, BinaryIO, Callable

from mrp.auth import Keypair, _b64url_decode
from mrp.client import Client
from mrp.crypto import decrypt_message, encrypt_message
from mrp.errors import RequestError
from mrp.models import (
    ACLEntry,
    AgentInfo,
    Attachment,
    Blob,
    CapabilitiesResult,
    DiscoveredAgent,
    Message,
    PollResult,
    SendResult,
)
from mrp.schemas import (
    REQUEST_CONTENT_TYPE,
    RESPONSE_CONTENT_TYPE,
    EVENT_CONTENT_TYPE,
    STATUS_CONTENT_TYPE,
    MRPRequest,
    create_request,
    create_ok_response,
    create_error_response,
    create_event,
    create_status,
    parse_request,
)
from mrp.websocket import ReconnectingWSClient

logger = logging.getLogger("mrp.agent")


class Agent:
    """High-level MRP agent with automatic lifecycle management."""

    def __init__(
        self,
        relay: str,
        key_file: str | None = None,
        keypair: Keypair | None = None,
        name: str | None = None,
        capabilities: list[str] | None = None,
        metadata: dict[str, str] | None = None,
        visibility: str | None = None,
        inbox_policy: str | None = None,
    ):
        if keypair is not None:
            self._keypair = keypair
        elif key_file is not None:
            self._keypair = Keypair.from_file(key_file, create=True)
        else:
            self._keypair = Keypair.generate()

        self._client = Client(relay, self._keypair)
        self._relay = relay
        self._name = name
        self._capabilities = capabilities
        self._metadata = metadata
        self._visibility = visibility
        self._inbox_policy = inbox_policy
        self._registered = False
        self._on_message_handler: Callable[[Message], dict | None] | None = None
        self._action_handlers: dict[str, Callable[[MRPRequest, Message], Any]] = {}

    @property
    def public_key(self) -> str:
        """Base64url public key string."""
        return self._keypair.public_key_b64

    @property
    def client(self) -> Client:
        """Access the underlying low-level client."""
        return self._client

    def _ensure_registered(self) -> None:
        """Register/update profile if not done yet."""
        if not self._registered:
            self.register()

    def register(self) -> AgentInfo:
        """Register/update profile on relay."""
        info = self._client.update_agent(
            self._keypair.public_key_b64,
            display_name=self._name,
            capabilities=self._capabilities,
            metadata=self._metadata,
            visibility=self._visibility,
            inbox_policy=self._inbox_policy,
        )
        self._registered = True
        return info

    # --- Discovery ---

    def discover(self, capability: str) -> list[DiscoveredAgent]:
        """Find agents by capability. Auto-paginates."""
        self._ensure_registered()
        return self._client.discover_all(capability)

    def discover_search(
        self,
        capabilities: list[str] | None = None,
        capability_prefix: str | None = None,
        name: str | None = None,
    ) -> list[DiscoveredAgent]:
        """Find agents by multiple criteria. Auto-paginates."""
        self._ensure_registered()
        return self._client.discover_search_all(
            capabilities=capabilities,
            capability_prefix=capability_prefix,
            name=name,
        )

    def list_capabilities(self) -> CapabilitiesResult:
        """List all capability tags with agent counts."""
        self._ensure_registered()
        return self._client.list_capabilities()

    # --- Messaging (sync, HTTP-based) ---

    def send(
        self,
        to: str,
        body: dict,
        content_type: str = "application/json",
        thread: str | None = None,
        in_reply_to: str | None = None,
        expires_in: int | None = None,
        attachments: list[str] | None = None,
        encrypt: bool = False,
    ) -> SendResult:
        """Send a message. If encrypt=True, E2E encrypt the body."""
        self._ensure_registered()

        actual_body: dict | None = body
        actual_ct = content_type

        if encrypt:
            recipient_pub = _b64url_decode(to)
            encrypted = encrypt_message(
                self._keypair.seed,
                recipient_pub,
                json.dumps(body).encode(),
                content_type,
            )
            actual_body = encrypted
            actual_ct = "application/x-m2m-encrypted"

        attach_list = None
        if attachments:
            attach_list = [Attachment(blob_id=bid) for bid in attachments]

        return self._client.send_message(
            recipient_key=to,
            body=actual_body,
            content_type=actual_ct,
            thread_id=thread,
            in_reply_to=in_reply_to,
            expires_in=expires_in,
            attachments=attach_list,
        )

    def reply(
        self,
        message: Message,
        body: dict,
        encrypt: bool = False,
    ) -> SendResult:
        """Reply to a message. Auto-sets in_reply_to and recipient."""
        return self.send(
            to=message.sender_key,
            body=body,
            in_reply_to=message.message_id,
            thread=message.thread_id,
            encrypt=encrypt,
        )

    # --- Structured Messaging ---

    def request(
        self,
        to: str,
        action: str,
        params: dict[str, Any],
        thread: str | None = None,
        response_format: str | None = None,
        encrypt: bool = False,
    ) -> SendResult:
        """Send a structured MRP request message."""
        body = create_request(action, params, response_format)
        return self.send(
            to=to,
            body=body.to_dict(),
            content_type=REQUEST_CONTENT_TYPE,
            thread=thread,
            encrypt=encrypt,
        )

    def respond(
        self,
        message: Message,
        result: dict[str, Any],
    ) -> SendResult:
        """Send an ok response to an MRP request message."""
        body = create_ok_response(result)
        return self.send(
            to=message.sender_key,
            body=body.to_dict(),
            content_type=RESPONSE_CONTENT_TYPE,
            in_reply_to=message.message_id,
            thread=message.thread_id,
        )

    def respond_error(
        self,
        message: Message,
        code: str,
        msg: str,
    ) -> SendResult:
        """Send an error response to an MRP request message."""
        body = create_error_response(code, msg)
        return self.send(
            to=message.sender_key,
            body=body.to_dict(),
            content_type=RESPONSE_CONTENT_TYPE,
            in_reply_to=message.message_id,
            thread=message.thread_id,
        )

    def emit_event(
        self,
        to: str,
        event: str,
        data: dict[str, Any] | None = None,
        thread: str | None = None,
        encrypt: bool = False,
    ) -> SendResult:
        """Send an MRP event message."""
        body = create_event(event, data)
        return self.send(
            to=to,
            body=body.to_dict(),
            content_type=EVENT_CONTENT_TYPE,
            thread=thread,
            encrypt=encrypt,
        )

    def send_status(
        self,
        to: str,
        progress: float | None = None,
        stage: str | None = None,
        detail: str | None = None,
        thread: str | None = None,
        encrypt: bool = False,
    ) -> SendResult:
        """Send an MRP status update message."""
        body = create_status(progress=progress, stage=stage, detail=detail)
        return self.send(
            to=to,
            body=body.to_dict(),
            content_type=STATUS_CONTENT_TYPE,
            thread=thread,
            encrypt=encrypt,
        )

    # --- Action Router ---

    def on_request(self, action: str, handler: Callable[[MRPRequest, Message], Any]) -> None:
        """Register a handler for a specific request action.

        The handler receives (MRPRequest, Message) and should return a dict.
        If the handler raises RequestError, an error response is auto-sent.
        """
        self._action_handlers[action] = handler

    def _dispatch_message(self, msg: Message) -> dict | None:
        """Dispatch a message through action routing then generic handler.

        Returns a reply dict if the generic handler produces one, else None.
        Action-routed handlers auto-send their own responses.
        """
        # Try action routing first for request messages
        if msg.content_type == REQUEST_CONTENT_TYPE and self._action_handlers:
            req = parse_request(msg.body)
            if req:
                handler = self._action_handlers.get(req.action)
                if handler:
                    try:
                        if inspect.iscoroutinefunction(handler):
                            result = asyncio.get_event_loop().run_until_complete(handler(req, msg))
                        else:
                            result = handler(req, msg)
                        self.respond(msg, result)
                    except RequestError as e:
                        self.respond_error(msg, e.code, e.message)
                    return None

        # Fall back to generic message handler
        if self._on_message_handler:
            return self._on_message_handler(msg)
        return None

    async def _dispatch_message_async(self, msg: Message) -> dict | None:
        """Async version of _dispatch_message for WebSocket mode."""
        # Try action routing first for request messages
        if msg.content_type == REQUEST_CONTENT_TYPE and self._action_handlers:
            req = parse_request(msg.body)
            if req:
                handler = self._action_handlers.get(req.action)
                if handler:
                    try:
                        if inspect.iscoroutinefunction(handler):
                            result = await handler(req, msg)
                        else:
                            result = handler(req, msg)
                        self.respond(msg, result)
                    except RequestError as e:
                        self.respond_error(msg, e.code, e.message)
                    return None

        # Fall back to generic message handler
        if self._on_message_handler:
            return self._on_message_handler(msg)
        return None

    def receive(
        self,
        limit: int = 50,
        sender: str | None = None,
    ) -> list[Message]:
        """One-shot poll for new messages."""
        self._ensure_registered()
        result = self._client.poll_messages(limit=limit, sender_key=sender)
        return result.messages

    def messages(
        self,
        poll_interval: float = 2.0,
        max_empty_interval: float = 30.0,
    ) -> Iterator[Message]:
        """Blocking iterator that polls for messages with smart backoff."""
        self._ensure_registered()
        interval = poll_interval

        while True:
            msgs = self.receive()
            if msgs:
                interval = poll_interval
                yield from msgs
            else:
                interval = min(interval * 2, max_empty_interval)
            time.sleep(interval)

    # --- Messaging (async, WebSocket-based) ---

    def on_message(self, handler: Callable[[Message], dict | None]) -> None:
        """Register a message handler for WebSocket mode."""
        self._on_message_handler = handler

    def run(self, *, mode: str = "websocket") -> None:
        """Block and process messages.

        mode="websocket": maintain WebSocket, dispatch to on_message handler.
        mode="poll": poll loop, dispatch to on_message handler.
        """
        self._ensure_registered()

        if mode == "poll":
            self._run_poll()
        elif mode == "websocket":
            asyncio.run(self._run_ws())
        else:
            raise ValueError(f"Unknown mode: {mode}")

    def _run_poll(self) -> None:
        """Poll loop that dispatches messages through action routing and on_message handler."""
        for msg in self.messages():
            reply = self._dispatch_message(msg)
            if reply is not None:
                self.reply(msg, reply)

    async def _run_ws(self) -> None:
        """WebSocket loop that dispatches messages through action routing and on_message handler."""
        ws = ReconnectingWSClient(self._relay, self._keypair)

        async def _dispatch(msg: Message) -> dict | None:
            return await self._dispatch_message_async(msg)

        ws.on_message(_dispatch)
        try:
            await ws.run()
        finally:
            await ws.close()

    # --- Blobs ---

    def upload(self, data: bytes | BinaryIO, content_type: str = "application/octet-stream") -> Blob:
        """Upload a blob."""
        self._ensure_registered()
        return self._client.upload_blob(data, content_type)

    def download(self, blob_id: str) -> tuple[bytes, str]:
        """Download a blob."""
        self._ensure_registered()
        return self._client.download_blob(blob_id)

    # --- Thread ---

    def thread(self, thread_id: str) -> list[Message]:
        """Get all messages in a thread (auto-paginates)."""
        self._ensure_registered()
        all_msgs: list[Message] = []
        cursor = None
        while True:
            result = self._client.get_thread(thread_id, cursor=cursor)
            all_msgs.extend(result.messages)
            if not result.next_cursor:
                break
            cursor = result.next_cursor
        return all_msgs

    # --- ACL ---

    def allow(self, peer_key: str) -> ACLEntry:
        """Allow a peer to message this agent."""
        self._ensure_registered()
        return self._client.set_acl(self._keypair.public_key_b64, peer_key, "allow")

    def block(self, peer_key: str) -> ACLEntry:
        """Block a peer from messaging this agent."""
        self._ensure_registered()
        return self._client.set_acl(self._keypair.public_key_b64, peer_key, "block")

    def unblock(self, peer_key: str) -> None:
        """Remove a peer from the ACL."""
        self._ensure_registered()
        self._client.delete_acl(self._keypair.public_key_b64, peer_key)

    def list_acl(self, entry_type: str | None = None) -> list[ACLEntry]:
        """List ACL entries, optionally filtered by type."""
        self._ensure_registered()
        return self._client.list_acl(self._keypair.public_key_b64, entry_type)

    # --- Lifecycle ---

    def delete(self) -> None:
        """Delete this agent from the relay."""
        self._client.delete_agent(self._keypair.public_key_b64)

    def close(self) -> None:
        """Close connections."""
        self._client.close()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    def decrypt(self, message: Message) -> tuple[bytes, str]:
        """Decrypt an E2E encrypted message."""
        if message.content_type != "application/x-m2m-encrypted":
            raise ValueError("Message is not encrypted")
        sender_pub = _b64url_decode(message.sender_key)
        return decrypt_message(self._keypair.seed, sender_pub, message.body)



import json
from pathlib import Path

import pytest
from mrp.auth import Keypair, _b64url_decode, _b64url_encode
from mrp.crypto import (
    decrypt_message,
    ed25519_to_x25519_private,
    ed25519_to_x25519_public,
    encrypt_message,
)


class TestKeyConversion:
    def test_ed25519_to_x25519_public(self):
        kp = Keypair.generate()
        x_pub = ed25519_to_x25519_public(kp.public_key)
        # X25519 public key should be 32 bytes
        from cryptography.hazmat.primitives import serialization
        raw = x_pub.public_bytes(serialization.Encoding.Raw, serialization.PublicFormat.Raw)
        assert len(raw) == 32

    def test_ed25519_to_x25519_private(self):
        kp = Keypair.generate()
        x_priv = ed25519_to_x25519_private(kp.seed)
        # Should produce a valid X25519 private key
        from cryptography.hazmat.primitives import serialization
        raw = x_priv.private_bytes(
            serialization.Encoding.Raw,
            serialization.PrivateFormat.Raw,
            serialization.NoEncryption(),
        )
        assert len(raw) == 32

    def test_different_keys_different_x25519(self):
        kp1 = Keypair.generate()
        kp2 = Keypair.generate()
        from cryptography.hazmat.primitives import serialization
        x1 = ed25519_to_x25519_public(kp1.public_key).public_bytes(
            serialization.Encoding.Raw, serialization.PublicFormat.Raw
        )
        x2 = ed25519_to_x25519_public(kp2.public_key).public_bytes(
            serialization.Encoding.Raw, serialization.PublicFormat.Raw
        )
        assert x1 != x2


class TestEncryptDecrypt:
    def test_round_trip(self):
        sender = Keypair.generate()
        recipient = Keypair.generate()
        plaintext = b'{"message": "hello"}'

        encrypted = encrypt_message(
            sender.seed,
            recipient.public_key,
            plaintext,
            "application/json",
        )

        # Verify encrypted body has v2 fields
        assert encrypted["v"] == 2
        assert "enc" in encrypted
        assert "ct" in encrypted
        assert encrypted["ct_content_type"] == "application/json"

        # Decrypt
        decrypted, ct = decrypt_message(
            recipient.seed,
            sender.public_key,
            encrypted,
        )
        assert decrypted == plaintext
        assert ct == "application/json"

    def test_decrypt_with_wrong_key_fails(self):
        sender = Keypair.generate()
        recipient = Keypair.generate()
        wrong = Keypair.generate()

        encrypted = encrypt_message(
            sender.seed,
            recipient.public_key,
            b"secret",
            "text/plain",
        )

        with pytest.raises(Exception):
            decrypt_message(wrong.seed, sender.public_key, encrypted)

    def test_different_encryptions_differ(self):
        sender = Keypair.generate()
        recipient = Keypair.generate()
        plaintext = b"same message"

        e1 = encrypt_message(sender.seed, recipient.public_key, plaintext)
        e2 = encrypt_message(sender.seed, recipient.public_key, plaintext)

        # Different enc values -> different ciphertexts
        assert e1["enc"] != e2["enc"]
        assert e1["ct"] != e2["ct"]

    def test_large_message(self):
        sender = Keypair.generate()
        recipient = Keypair.generate()
        plaintext = b"x" * 100_000

        encrypted = encrypt_message(sender.seed, recipient.public_key, plaintext)
        decrypted, _ = decrypt_message(recipient.seed, sender.public_key, encrypted)
        assert decrypted == plaintext

    def test_empty_message(self):
        sender = Keypair.generate()
        recipient = Keypair.generate()

        encrypted = encrypt_message(sender.seed, recipient.public_key, b"")
        decrypted, _ = decrypt_message(recipient.seed, sender.public_key, encrypted)
        assert decrypted == b""

    def test_unsupported_version(self):
        sender = Keypair.generate()
        recipient = Keypair.generate()

        encrypted = encrypt_message(sender.seed, recipient.public_key, b"test")
        encrypted["v"] = 99

        with pytest.raises(ValueError, match="Unsupported encryption version"):
            decrypt_message(recipient.seed, sender.public_key, encrypted)


class TestCrossSDKVectors:
    @pytest.fixture
    def fixture(self):
        fixture_path = Path(__file__).parent / ".." / ".." / "testdata" / "crypto_test_vectors.json"
        return json.loads(fixture_path.read_text())

    def test_decrypt_basic_json(self, fixture):
        self._decrypt_vector(fixture, "basic_json")

    def test_decrypt_empty_message(self, fixture):
        self._decrypt_vector(fixture, "empty_message")

    def test_decrypt_binary_payload(self, fixture):
        self._decrypt_vector(fixture, "binary_payload")

    def test_wrong_sender_fails(self, fixture):
        """Decryption must fail with a wrong sender key."""
        recipient_seed = bytes.fromhex(fixture["recipient_seed_hex"])
        wrong = Keypair.generate()

        for vec in fixture["vectors"]:
            with pytest.raises(Exception):
                decrypt_message(
                    recipient_seed, wrong.public_key, vec["encrypted_body"],
                )

    def _decrypt_vector(self, fixture, name):
        recipient_seed = bytes.fromhex(fixture["recipient_seed_hex"])
        sender_pub = _b64url_decode(fixture["sender_public_key_b64url"])

        vec = next(v for v in fixture["vectors"] if v["name"] == name)
        expected_plaintext = _b64url_decode(vec["plaintext_b64url"]) if vec["plaintext_b64url"] else b""

        decrypted, ct = decrypt_message(
            recipient_seed, sender_pub, vec["encrypted_body"],
        )
        assert decrypted == expected_plaintext
        assert ct == vec["expected_content_type"]


class TestCiphertextTampering:
    def test_tampered_ciphertext(self):
        sender = Keypair.generate()
        recipient = Keypair.generate()
        encrypted = encrypt_message(sender.seed, recipient.public_key, b"secret")
        ct_bytes = bytearray(_b64url_decode(encrypted["ct"]))
        ct_bytes[0] ^= 0xFF
        encrypted["ct"] = _b64url_encode(bytes(ct_bytes))
        with pytest.raises(Exception):
            decrypt_message(recipient.seed, sender.public_key, encrypted)

    def test_tampered_enc(self):
        sender = Keypair.generate()
        recipient = Keypair.generate()
        encrypted = encrypt_message(sender.seed, recipient.public_key, b"secret")
        enc_bytes = bytearray(_b64url_decode(encrypted["enc"]))
        enc_bytes[0] ^= 0xFF
        encrypted["enc"] = _b64url_encode(bytes(enc_bytes))
        with pytest.raises(Exception):
            decrypt_message(recipient.seed, sender.public_key, encrypted)


class TestMalformedFields:
    def test_invalid_base64_enc(self):
        sender = Keypair.generate()
        recipient = Keypair.generate()
        encrypted = encrypt_message(sender.seed, recipient.public_key, b"test")
        encrypted["enc"] = "!!!invalid!!!"
        with pytest.raises(Exception):
            decrypt_message(recipient.seed, sender.public_key, encrypted)

    def test_invalid_base64_ct(self):
        sender = Keypair.generate()
        recipient = Keypair.generate()
        encrypted = encrypt_message(sender.seed, recipient.public_key, b"test")
        encrypted["ct"] = "!!!invalid!!!"
        with pytest.raises(Exception):
            decrypt_message(recipient.seed, sender.public_key, encrypted)

    def test_missing_ct_content_type(self):
        sender = Keypair.generate()
        recipient = Keypair.generate()
        encrypted = encrypt_message(sender.seed, recipient.public_key, b"test")
        del encrypted["ct_content_type"]
        with pytest.raises(KeyError):
            decrypt_message(recipient.seed, sender.public_key, encrypted)


class TestSelfEncrypt:
    def test_self_encrypt_decrypt(self):
        """Sender == recipient is valid (agent encrypting to itself for storage)."""
        kp = Keypair.generate()
        plaintext = b'{"self": "message"}'

        encrypted = encrypt_message(kp.seed, kp.public_key, plaintext, "application/json")
        decrypted, ct = decrypt_message(kp.seed, kp.public_key, encrypted)
        assert decrypted == plaintext
        assert ct == "application/json"


class TestSenderKeyRequired:
    def test_decrypt_fails_with_wrong_sender_key(self):
        """In v2 (HPKE Auth mode), the sender's key is cryptographically bound.
        Decryption with a wrong sender public key must fail."""
        sender = Keypair.generate()
        recipient = Keypair.generate()
        unrelated = Keypair.generate()

        encrypted = encrypt_message(sender.seed, recipient.public_key, b"hello")

        with pytest.raises(Exception):
            decrypt_message(recipient.seed, unrelated.public_key, encrypted)



from __future__ import annotations

import json
import tempfile
from pathlib import Path

import pytest

from mrp.contacts import Contact, ContactStore

GOLDEN_FILE = str(Path(__file__).parent / "../../testdata/contacts_golden.json")


class TestGoldenFile:
    """Cross-SDK interoperability: verify we can load the shared golden file."""

    def test_loads_golden_file(self):
        store = ContactStore.load(GOLDEN_FILE)
        assert len(store.list()) == 2

        alice = store.get("Alice")
        assert alice is not None
        assert alice.public_key == "ed25519:O2onvM62pC1io6jQKm8Nc2UyFXcd4kOmOsBIoYtZ2ik"
        assert alice.note == "translation agent"

        bob = store.get("Bob")
        assert bob is not None
        assert bob.public_key == "ed25519:Xk3m9nR4pQ7sT2wV5yA8bC1dE6fG0hI3jK4lM7nO9pQ"
        assert bob.note is None

        # Verify resolve works
        assert store.resolve("Alice") == alice.public_key
        assert store.resolve("bob") == bob.public_key


class TestDefaultPath:
    def test_returns_absolute_path(self):
        p = ContactStore.default_path()
        assert Path(p).is_absolute()
        assert p.endswith("contacts.json")


class TestLoad:
    def test_no_file(self, tmp_path):
        store = ContactStore.load(str(tmp_path / "missing.json"))
        assert store.list() == []

    def test_invalid_json(self, tmp_path):
        p = tmp_path / "bad.json"
        p.write_text("{bad}")
        with pytest.raises(ValueError, match="parsing contacts"):
            ContactStore.load(str(p))

    def test_loads_existing(self, tmp_path):
        p = tmp_path / "contacts.json"
        p.write_text(
            json.dumps(
                {
                    "contacts": [
                        {"name": "Alice", "public_key": "ed25519:alice", "note": "friend"},
                    ]
                }
            )
        )
        store = ContactStore.load(str(p))
        assert len(store.list()) == 1
        assert store.list()[0].name == "Alice"
        assert store.list()[0].public_key == "ed25519:alice"
        assert store.list()[0].note == "friend"


class TestSaveAndLoad:
    def test_round_trip(self, tmp_path):
        p = str(tmp_path / "contacts.json")
        store = ContactStore.load(p)
        store.add("Alice", "ed25519:alice-key", "test agent")
        store.add("Bob", "ed25519:bob-key")
        store.save()

        loaded = ContactStore.load(p)
        assert len(loaded.list()) == 2
        assert loaded.list()[0].name == "Alice"
        assert loaded.list()[0].note == "test agent"
        assert loaded.list()[1].name == "Bob"
        assert loaded.list()[1].note is None


class TestAdd:
    def test_adds_contact(self):
        store = ContactStore("/tmp/unused.json")
        store.add("Alice", "ed25519:key1")
        assert len(store.list()) == 1

    def test_duplicate_name(self):
        store = ContactStore("/tmp/unused.json")
        store.add("Alice", "ed25519:key1")
        with pytest.raises(ValueError, match="already exists"):
            store.add("alice", "ed25519:key2")

    def test_duplicate_key(self):
        store = ContactStore("/tmp/unused.json")
        store.add("Alice", "ed25519:key1")
        with pytest.raises(ValueError, match="already saved"):
            store.add("Bob", "ed25519:key1")

    def test_empty_name(self):
        store = ContactStore("/tmp/unused.json")
        with pytest.raises(ValueError, match="must not be empty"):
            store.add("", "ed25519:key1")

    def test_name_with_colon(self):
        store = ContactStore("/tmp/unused.json")
        with pytest.raises(ValueError, match="must not contain"):
            store.add("bad:name", "ed25519:key1")

    def test_name_too_long(self):
        store = ContactStore("/tmp/unused.json")
        with pytest.raises(ValueError, match="at most 64"):
            store.add("a" * 65, "ed25519:key1")


class TestRemove:
    def test_removes_case_insensitive(self):
        store = ContactStore("/tmp/unused.json")
        store.add("Alice", "ed25519:key1")
        store.add("Bob", "ed25519:key2")
        store.remove("alice")
        assert len(store.list()) == 1
        assert store.list()[0].name == "Bob"

    def test_not_found(self):
        store = ContactStore("/tmp/unused.json")
        with pytest.raises(ValueError, match="not found"):
            store.remove("nobody")


class TestResolve:
    def test_found_by_name(self):
        store = ContactStore("/tmp/unused.json")
        store.add("Alice", "ed25519:alice-key")
        assert store.resolve("Alice") == "ed25519:alice-key"

    def test_case_insensitive(self):
        store = ContactStore("/tmp/unused.json")
        store.add("Alice", "ed25519:alice-key")
        assert store.resolve("alice") == "ed25519:alice-key"

    def test_not_found_returns_input(self):
        store = ContactStore("/tmp/unused.json")
        assert store.resolve("nobody") == "nobody"

    def test_raw_key_passthrough(self):
        store = ContactStore("/tmp/unused.json")
        assert store.resolve("ed25519:raw-key") == "ed25519:raw-key"


class TestGet:
    def test_finds_by_name(self):
        store = ContactStore("/tmp/unused.json")
        store.add("Alice", "ed25519:key1", "note")
        c = store.get("alice")
        assert c is not None
        assert c.public_key == "ed25519:key1"

    def test_not_found(self):
        store = ContactStore("/tmp/unused.json")
        assert store.get("nobody") is None


class TestList:
    def test_returns_all(self):
        store = ContactStore("/tmp/unused.json")
        store.add("Alice", "ed25519:key1")
        store.add("Bob", "ed25519:key2")
        assert len(store.list()) == 2



import asyncio
import json

import pytest
import websockets

from unittest.mock import AsyncMock, MagicMock, patch

from mrp.auth import Keypair
from mrp.websocket import WSClient, ReconnectingWSClient


class TestWSUrl:
    def test_http_to_ws(self):
        ws = WSClient("http://localhost:8080", Keypair.generate())
        assert ws._ws_url() == "ws://localhost:8080/v1/ws"

    def test_https_to_wss(self):
        ws = WSClient("https://relay.example.com", Keypair.generate())
        assert ws._ws_url() == "wss://relay.example.com/v1/ws"

    def test_trailing_slash_stripped(self):
        ws = WSClient("http://localhost:8080/", Keypair.generate())
        assert ws._ws_url() == "ws://localhost:8080/v1/ws"


def _make_mock_conn(recv_queue: asyncio.Queue):
    """Create a mock WebSocket connection backed by an asyncio.Queue."""
    mock_conn = AsyncMock()
    mock_conn.send = AsyncMock()
    mock_conn.close = AsyncMock()

    async def mock_recv():
        return await recv_queue.get()

    mock_conn.recv = mock_recv

    async def mock_aiter(self):
        while True:
            try:
                data = await asyncio.wait_for(recv_queue.get(), timeout=0.5)
                yield data
            except asyncio.TimeoutError:
                break

    mock_conn.__aiter__ = mock_aiter
    return mock_conn


class TestWSConnect:
    @pytest.mark.asyncio
    async def test_auth_handshake(self):
        """Verify the auth frame is sent with required fields and auth_result is consumed."""
        recv_queue = asyncio.Queue()
        mock_conn = _make_mock_conn(recv_queue)

        # Queue the auth_result response
        await recv_queue.put(json.dumps({"type": "auth_result", "status": "ok"}))

        keypair = Keypair.generate()
        ws = WSClient("http://localhost:8080", keypair)

        mock_connect = AsyncMock(return_value=mock_conn)
        with patch("mrp.websocket.websockets.connect", mock_connect):
            await ws.connect()

        # Verify auth frame was sent
        assert mock_conn.send.call_count == 1
        auth_frame = json.loads(mock_conn.send.call_args[0][0])
        assert auth_frame["type"] == "auth"
        assert auth_frame["public_key"] == keypair.public_key_b64
        assert "timestamp" in auth_frame
        assert "signature" in auth_frame

    @pytest.mark.asyncio
    async def test_auth_failure(self):
        """Server returns auth error -- should raise ConnectionError."""
        recv_queue = asyncio.Queue()
        mock_conn = _make_mock_conn(recv_queue)

        await recv_queue.put(json.dumps({
            "type": "auth_result", "status": "error", "error": "bad"
        }))

        keypair = Keypair.generate()
        ws = WSClient("http://localhost:8080", keypair)

        mock_connect = AsyncMock(return_value=mock_conn)
        with patch("mrp.websocket.websockets.connect", mock_connect):
            with pytest.raises(ConnectionError, match="Auth failed"):
                await ws.connect()

    @pytest.mark.asyncio
    async def test_auth_timeout(self):
        """recv never returns -- should raise ConnectionError after timeout."""
        recv_queue = asyncio.Queue()
        mock_conn = _make_mock_conn(recv_queue)

        # Override recv to hang
        async def slow_recv():
            await asyncio.sleep(60)

        mock_conn.recv = slow_recv

        keypair = Keypair.generate()
        ws = WSClient("http://localhost:8080", keypair)

        mock_connect = AsyncMock(return_value=mock_conn)
        with patch("mrp.websocket.websockets.connect", mock_connect):
            # Patch the timeout to be very short so the test is fast
            with patch("mrp.websocket.asyncio.wait_for", side_effect=asyncio.TimeoutError):
                with pytest.raises(ConnectionError, match="Auth response timeout"):
                    await ws.connect()


class TestWSMessageLoop:
    @pytest.mark.asyncio
    async def test_ping_pong(self):
        """Put a ping frame in the queue and verify pong is sent back."""
        recv_queue = asyncio.Queue()
        mock_conn = _make_mock_conn(recv_queue)

        await recv_queue.put(json.dumps({"type": "ping"}))

        keypair = Keypair.generate()
        ws = WSClient("http://localhost:8080", keypair)
        ws._ws = mock_conn

        await ws._read_loop()

        # Find the pong send call
        pong_calls = [
            call for call in mock_conn.send.call_args_list
            if json.loads(call[0][0]).get("type") == "pong"
        ]
        assert len(pong_calls) == 1

    @pytest.mark.asyncio
    async def test_receive_message_and_ack(self):
        """Put a message frame in the queue, verify on_message called and ack sent."""
        recv_queue = asyncio.Queue()
        mock_conn = _make_mock_conn(recv_queue)

        msg_frame = {
            "type": "message",
            "message": {
                "message_id": "msg_1",
                "sender_key": "sender_abc",
                "recipient_key": "recipient_xyz",
                "content_type": "application/json",
                "body": {"text": "hello"},
                "attachments": [],
                "status": "delivered",
                "metadata": {},
                "created_at": "2026-03-05T12:00:00Z",
                "expires_at": "2026-03-12T12:00:00Z",
            },
        }
        await recv_queue.put(json.dumps(msg_frame))

        keypair = Keypair.generate()
        ws = WSClient("http://localhost:8080", keypair)
        ws._ws = mock_conn

        received_messages = []

        async def handler(msg):
            received_messages.append(msg)
            return None

        ws.on_message(handler)
        await ws._read_loop()

        # Verify handler was called
        assert len(received_messages) == 1
        assert received_messages[0].message_id == "msg_1"
        assert received_messages[0].body == {"text": "hello"}

        # Verify ack was sent
        ack_calls = [
            call for call in mock_conn.send.call_args_list
            if json.loads(call[0][0]).get("type") == "ack"
        ]
        assert len(ack_calls) == 1
        ack_frame = json.loads(ack_calls[0][0][0])
        assert ack_frame["message_id"] == "msg_1"

    @pytest.mark.asyncio
    async def test_send_result_resolves(self):
        """Create a pending send, put send_result in queue, verify future resolved."""
        recv_queue = asyncio.Queue()
        mock_conn = _make_mock_conn(recv_queue)

        request_id = "req_123"
        send_result_frame = {
            "type": "send_result",
            "request_id": request_id,
            "status": "ok",
            "message_id": "msg_99",
        }
        await recv_queue.put(json.dumps(send_result_frame))

        keypair = Keypair.generate()
        ws = WSClient("http://localhost:8080", keypair)
        ws._ws = mock_conn

        # Create a pending future
        loop = asyncio.get_event_loop()
        future = loop.create_future()
        ws._pending_sends[request_id] = future

        await ws._read_loop()

        # Future should be resolved
        assert future.done()
        result = future.result()
        assert result["status"] == "ok"
        assert result["message_id"] == "msg_99"


class TestWSSendMessage:
    @pytest.mark.asyncio
    async def test_send_success(self):
        """Mock ws connected, call send_message, resolve the pending future, verify frame."""
        recv_queue = asyncio.Queue()
        mock_conn = _make_mock_conn(recv_queue)
        mock_conn.send = AsyncMock()

        keypair = Keypair.generate()
        ws = WSClient("http://localhost:8080", keypair)
        ws._ws = mock_conn

        async def resolve_send():
            # Wait for the pending send to be registered
            while not ws._pending_sends:
                await asyncio.sleep(0.01)
            # Resolve it
            for req_id, future in ws._pending_sends.items():
                if not future.done():
                    future.set_result({
                        "status": "ok",
                        "message_id": "msg_42",
                        "request_id": req_id,
                    })

        # Run send_message and the resolver concurrently
        result, _ = await asyncio.gather(
            ws.send_message("recipient_key_abc", {"text": "hi"}),
            resolve_send(),
        )

        assert result.message_id == "msg_42"
        assert result.status == "sent"

        # Verify the sent frame structure
        sent_data = json.loads(mock_conn.send.call_args[0][0])
        assert sent_data["type"] == "send"
        assert sent_data["recipient_key"] == "recipient_key_abc"
        assert sent_data["body"] == {"text": "hi"}
        assert "request_id" in sent_data

    @pytest.mark.asyncio
    async def test_send_not_connected(self):
        """ws is None -- should raise ConnectionError."""
        keypair = Keypair.generate()
        ws = WSClient("http://localhost:8080", keypair)
        # ws._ws is None by default

        with pytest.raises(ConnectionError, match="Not connected"):
            await ws.send_message("recipient_key", {"text": "hi"})


class TestReconnectingWSClient:
    @pytest.mark.asyncio
    async def test_reconnects_on_disconnect(self):
        """First connect succeeds then read_loop raises ConnectionClosed, verify reconnection."""
        keypair = Keypair.generate()
        ws = ReconnectingWSClient(
            "http://localhost:8080", keypair,
            max_retries=3, base_delay=0.01, max_delay=0.05,
        )

        connect_count = 0

        async def mock_connect():
            nonlocal connect_count
            connect_count += 1

        call_count = 0

        async def mock_read_loop():
            nonlocal call_count
            call_count += 1
            if call_count == 1:
                raise websockets.ConnectionClosed(None, None)
            else:
                # Second time, close cleanly to exit the loop
                ws._closed = True

        with patch.object(ws, "connect", side_effect=mock_connect):
            with patch.object(ws, "_read_loop", side_effect=mock_read_loop):
                await ws.run()

        assert connect_count >= 2

    @pytest.mark.asyncio
    async def test_max_retries(self):
        """connect always fails -- should stop after max_retries."""
        keypair = Keypair.generate()
        max_retries = 3
        ws = ReconnectingWSClient(
            "http://localhost:8080", keypair,
            max_retries=max_retries, base_delay=0.01, max_delay=0.05,
        )

        connect_count = 0

        async def mock_connect():
            nonlocal connect_count
            connect_count += 1
            raise ConnectionError("refused")

        with patch.object(ws, "connect", side_effect=mock_connect):
            await ws.run()

        assert connect_count == max_retries

    @pytest.mark.asyncio
    async def test_close_stops(self):
        """Call close() during run, verify loop ends."""
        keypair = Keypair.generate()
        ws = ReconnectingWSClient(
            "http://localhost:8080", keypair,
            max_retries=10, base_delay=0.01, max_delay=0.05,
        )

        async def mock_connect():
            pass

        async def mock_read_loop():
            # Simulate running for a bit, then close is called externally
            await asyncio.sleep(0.05)
            raise websockets.ConnectionClosed(None, None)

        async def close_after_delay():
            await asyncio.sleep(0.02)
            await ws.close()

        with patch.object(ws, "connect", side_effect=mock_connect):
            with patch.object(ws, "_read_loop", side_effect=mock_read_loop):
                # Run both concurrently -- close should cause run to exit
                await asyncio.gather(ws.run(), close_after_delay())

        assert ws._closed is True
